local Grid = require 'grid'
local Stack = require 'stack'
local Cursor = require 'cursor'
local Stream = require 'stream'

Befunge = {
  run = false,
  symb = false,
  task = nil,
  grid = nil,
  stack = nil,
  ip = nil
}
Befunge.__index = Befunge

-- Utils

local function __ternary(cond, a, b)
  if (cond) then
    return a
  else 
    return b
  end
end

local function __in(tab, val)
  for index, value in ipairs(tab) do
      if value == val then
          return true
      end
  end
  return false
end

-- Public methods

function Befunge.new(width, height, task)
  local bfi = {}
  setmetatable(bfi, Befunge)
  bfi.grid = Grid.new(width, height)
  bfi.stack = Stack.new()
  bfi.ip = Cursor.new(0, 0, bfi.grid.width, bfi.grid.height, DIR_RIGHT)
  bfi.task = task
  return bfi
end

function Befunge:start()
  self.run = true
  self.stack = Stack.new()
  self.ip = Cursor.new(0, 0, self.grid.width, self.grid.height, DIR_RIGHT)
  self.task:reset()
end

function Befunge:step()
  if (self.run) then
    local x = self.ip.x
    local y = self.ip.y
    local c = self.grid:get(x,y)

    if (c == "@") then
      self.run = false
    elseif (c == "\"") then
      self.symb = not self.symb
    elseif (Befunge.isNumeric(c)) then
      self.stack:push(tonumber(c))
    elseif (Befunge.isNullary(c)) then 
      self:nullary(c)
    elseif (Befunge.isUnary(c)) then
      self:unary(c)
    elseif (Befunge.isBinary(c)) then
      self:binary(c)
    else
      if (self.symb) then
        self.stack:push(string.byte(c))
      end
    end

    self.ip:move()
  end
end

function Befunge:stop()
  self.run = false
end

function Befunge.isNumeric(char)
  return tonumber(char) ~= nil
end

function Befunge.isNullary(char)
  return __in({"<", ">", "^", "v", "?", "#", "&"}, char)
end

function Befunge.isBinary(char)
  return __in({"+", "-", "*", "/", "%", "`", "\\"}, char)
end

function Befunge.isUnary(char)
  return __in({"_", "|", ":", "$", "!", ".", ","}, char)
end

-- Private methods

function Befunge:nullary(c)
  if (c == ">") then
    self.ip:setDirection(DIR_RIGHT)
  elseif (c == "<") then 
    self.ip:setDirection(DIR_LEFT)
  elseif (c == "^") then
    self.ip:setDirection(DIR_UP)
  elseif (c == "v") then
    self.ip:setDirection(DIR_DOWN)
  elseif (c == "?") then 
    self.ip:setRandomDirection()
  elseif (c == "#") then 
    self.ip:move()
  elseif (c == "&") then
    -- Check stream position
    local t = self.task.input:get()
    print(t)
    self.stack:push(t)
  end
end


function Befunge:unary(c)
  -- Check stack size
  local a = self.stack:pop()

  if (c == "_") then
    self.ip:setDirection(__ternary(a == 0, DIR_RIGHT, DIR_LEFT))
  elseif (c == "|") then 
    self.ip:setDirection(__ternary(a == 0, DIR_DOWN, DIR_UP))
  elseif (c == ":") then
    self.ip:push(a)
    self.ip:push(a)
  elseif (c == "$") then
    -- Nothing
  elseif (c == "!") then
    self.stack:push(__ternary(a == 0, 1, 0))
  elseif (c == ".") then
    -- print a as number
    print(a)
  elseif (c == ",") then
    -- print a as char (temp!!!!!)
    print(string.char(a))
  end
end

function Befunge:binary(c)
  -- Check stack size
  local b = self.stack:pop()
  local a = self.stack:pop()

  if (c == "+") then 
    self.stack:push(a + b)
  elseif (c == "-") then 
    self.stack:push(a - b)
  elseif (c == "*") then
    self.stack:push(a * b)
  elseif (c == "/") then 
    -- Check division by zero
    self.stack:push(a / b)
  elseif (c == "%") then
    -- Check division by zero
    self.stack:push(a % b)
  elseif (c == "\\") then
    self.stack:push(b)
    self.stack:push(a)
  elseif (c == "`") then
    self.stack:push(__ternary(a > b, 1, 0))
  end
end

return Befunge;